/*
 *  PlayStation 2 Remote Controller driver
 *
 *        Copyright (C) 2000, 2001  Sony Computer Entertainment Inc.
 *
 *  This file is subject to the terms and conditions of the GNU General
 *  Public License Version 2. See the file "COPYING" in the main
 *  directory of this archive for more details.
 *
 * $Id: rm.c,v 1.2 2001/11/20 08:49:37 takemura Exp $
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/malloc.h>
#include <linux/major.h>
#include <asm/addrspace.h>
#include <asm/uaccess.h>
#include <asm/ps2/siflock.h>

#include "rmcall.h"

#define PORT(n)		(((n) & 0x10) >> 4)
#define SLOT(n)		(((n) & 0x0f) >> 0)
#define NPORTS		2
#define NSLOTS		1	/* currently, we doesn't support multitap */

struct ps2rm_dev {
	int port, slot;
	int opened;
};

static ssize_t ps2rm_read(struct file *, char *, size_t, loff_t *);
static unsigned int ps2rm_poll(struct file *file, poll_table * wait);
static int ps2rm_ioctl(struct inode *, struct file *, u_int, u_long);
static int ps2rm_open(struct inode *, struct file *);
static int ps2rm_release(struct inode *, struct file *);

static int ps2rm_major = PS2RM_MAJOR;
ps2sif_lock_t *ps2rm_lock;	/* the lock which is need to call SBIOS */
static struct ps2rm_dev ps2rm_devs[NPORTS][NSLOTS];

#ifdef MODULE
EXPORT_NO_SYMBOLS;
MODULE_PARM(ps2rm_major, "0-255i");
#endif

#define PS2RM_DEBUG
#ifdef PS2RM_DEBUG
int ps2rm_debug = 0;
#define DPRINT(fmt, args...) \
	if (ps2rm_debug) printk(KERN_CRIT "ps2rm: " fmt, ## args)
#ifdef MODULE
MODULE_PARM(ps2rm_debug, "0-1i");
#endif
#else
#define DPRINT(fmt, args...)
#endif

#define ARRAYSIZEOF(a)	(sizeof(a)/sizeof(*(a)))

static struct file_operations ps2rm_fops = {
	NULL,		/* lseek	*/
	ps2rm_read,
	NULL,		/* write	*/
	NULL,		/* readdir	*/
	ps2rm_poll,
	ps2rm_ioctl,
	NULL,		/* mmap		*/
	ps2rm_open,
	NULL,		/* flush	*/
	ps2rm_release,
};

static ssize_t
ps2rm_read(struct file *filp, char *buf, size_t size, loff_t *off)
{
	int res;
	struct ps2rm_dev *dev = filp->private_data;
	u_char data[SB_REMOCON_MAXDATASIZE];

	if ((res = ps2sif_lock(ps2rm_lock, "read")) < 0)
		return (res);
	res = ps2rmlib_Read(dev->port, dev->slot, data, sizeof(data));
	ps2sif_unlock(ps2rm_lock);
	if (res <= 0)
		return -EIO;	/* error */

	if (size < res)
		res = size;

	if (copy_to_user(buf, data, res))
		return -EFAULT;

	return (res);
}

static int
ps2rm_ioctl(struct inode *inode, struct file *filp, u_int cmd, u_long arg)
{
	return -EINVAL;
}

static int
ps2rm_open(struct inode *inode, struct file *filp)
{
	int res;
	kdev_t devno = inode->i_rdev;
	struct ps2rm_dev *dev;
	int port, slot;

	/* diagnosis */
	if (MAJOR(devno) != ps2rm_major) {
		printk(KERN_ERR "ps2rm: incorrect major no\n");
		return -ENODEV;
	}

	port = PORT(devno);
	slot = SLOT(devno);

	DPRINT("open(%d, %d) devno=%04x\n", port, slot, devno);

	if (port < 0 || NPORTS <= port ||
	    slot < 0 || NSLOTS <= slot) {
		printk(KERN_ERR "ps2rm: invalid port or slot\n");
		return -ENODEV;
	}

	if ((res = ps2sif_lock(ps2rm_lock, "open")) < 0)
		return (res);
	dev = &ps2rm_devs[port][slot];
	DPRINT("open, dev=%lx %s\n", (unsigned long)dev,
	       dev->opened ? "already opened" : "");
	if (!dev->opened) {
		if (ps2rmlib_PortOpen(port, slot) != 1) {
			ps2sif_unlock(ps2rm_lock);
			DPRINT("ps2rmlib_PortOpen() failed\n");
			return -EIO;
		}
		dev->opened = 1;
	}
	ps2sif_unlock(ps2rm_lock);
	filp->private_data = dev;

	MOD_INC_USE_COUNT;

	return (0);
}

static unsigned int
ps2rm_poll(struct file *file, poll_table * wait)
{
	return POLLIN | POLLRDNORM;
}

static int
ps2rm_release(struct inode *inode, struct file *filp)
{
	int res;
	struct ps2rm_dev *dev = filp->private_data;

	DPRINT("close, dev=%lx\n", (unsigned long)dev);

	/* XXX, ignore errors */
	if ((res = ps2sif_lock(ps2rm_lock, "release")) == 0) {
		if (ps2rmlib_PortClose(dev->port, dev->slot) != 1)
			printk(KERN_ERR "ps2rm: ps2rmlib_PortClose() failed\n");
		dev->opened = 0;
		ps2sif_unlock(ps2rm_lock);
	} else {
		dev->opened = 0;
	}

	MOD_DEC_USE_COUNT;

	return 0;
}

int ps2rm_init(void)
{
	int res;
	int port, slot;

	DPRINT("PlayStation 2 remote controller: initialize...\n");

	if ((ps2rm_lock = ps2sif_getlock(PS2LOCK_REMOCON)) == NULL) {
		printk(KERN_ERR "ps2rm: Can't get lock\n");
		return -EINVAL;
	}

	if ((res = ps2sif_lock(ps2rm_lock, "initialize")) < 0) {
		printk(KERN_ERR "ps2rm: Can't lock\n");
		return (res);
	}

	/*
	 * initialize library
	 */
	if (ps2rmlib_Init(SBR_REMOCON_INIT_MODE) != 1) {
		printk(KERN_ERR "ps2rm: failed to initialize\n");
		ps2sif_unlock(ps2rm_lock);
		return -EIO;
	}
	ps2sif_unlock(ps2rm_lock);

	for (port = 0; port < NPORTS; port++) {
		for (slot = 0; slot < NSLOTS; slot++) {
			ps2rm_devs[port][slot].port = port;
			ps2rm_devs[port][slot].slot = slot;
		}
	}

	/*
	 * register device entry
	 */
	if ((res = register_chrdev(ps2rm_major, "ps2rm", &ps2rm_fops)) < 0) {
		printk(KERN_ERR "ps2rm: can't get major %d\n", ps2rm_major);
		return res;
	}
	if (ps2rm_major == 0)
		ps2rm_major = res;

	return (0);
}

#ifdef MODULE
int
init_module(void)
{
	DPRINT("load\n");
	return ps2rm_init();
}

void
cleanup_module(void)
{
#ifndef PS2PAD_NOPORTCLOSE
	int res, i;
#endif

	DPRINT("unload\n");

	if (unregister_chrdev(ps2rm_major, "ps2rm") < 0) {
		printk(KERN_WARNING "ps2rm: unregister_chrdev() error\n");
	}

	if ((res = ps2sif_lock(ps2rm_lock, "end")) < 0) {
		printk(KERN_ERR "ps2rm: Can't lock\n");
		return;
	}

	/*
	 * un-load library
	 */
	if (ps2rmlib_End() != 1)
		printk(KERN_ERR "ps2rm: ps2rmlib_End() failed\n");
	ps2sif_unlock(ps2rm_lock);
}
#endif /* MODULE */
